Package: DEoptimR
Version: 1.2-0
Date: 2026-06-07
Title: Differential Evolution Optimization in Pure R
Authors@R: c(
    person(c("Eduardo", "L. T."), "Conceicao",
           role = c("aut", "cre"),
           email = "mail@eduardoconceicao.org"),
    person("Martin", "Maechler",
           role = "ctb",
           email = "maechler@stat.math.ethz.ch",
           comment = c(ORCID = "0000-0002-8685-9910"))
    )
URL: svn://svn.r-forge.r-project.org/svnroot/robustbase/pkg/DEoptimR
Description: Differential Evolution (DE) stochastic heuristic algorithms for
    global optimization of problems with and without general constraints.
    The aim is to curate a collection of its variants that
    (1) do not sacrifice simplicity of design,
    (2) are essentially tuning-free, and
    (3) can be efficiently implemented directly in the R language.
    Currently, it provides implementations of the algorithms 'jDE' by
    Brest et al. (2006) <doi:10.1109/TEVC.2006.872133> for single-objective
    optimization and 'NCDE' by Qu et al. (2012) <doi:10.1109/TEVC.2011.2161873>
    for multimodal optimization (single-objective problems with
    multiple solutions).
Imports: stats, methods
Suggests: mirai (>= 2.7.1)
Enhances: robustbase
License: GPL (>= 2)
Author: Eduardo L. T. Conceicao [aut, cre],
  Martin Maechler [ctb] (ORCID: <https://orcid.org/0000-0002-8685-9910>)
Maintainer: Eduardo L. T. Conceicao <mail@eduardoconceicao.org>
Repository: CRAN
Repository/R-Forge/Project: robustbase
Repository/R-Forge/Revision: 1025
Repository/R-Forge/DateTimeStamp: 2026-06-07 16:58:58
Date/Publication: 2026-06-07 18:20:07 UTC
NeedsCompilation: no
Packaged: 2026-06-07 17:03:03 UTC; rforge
Built: R 4.4.3; ; 2026-06-07 21:42:13 UTC; unix
